<?php

namespace Drp\UserManagement;

use Drp\UserManagement\Contracts\ManagedUser;
use Drp\UserManagement\Events\AccountActivated;
use Illuminate\Auth\Authenticatable;
use Illuminate\Auth\Passwords\CanResetPassword;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Foundation\Auth\Access\Authorizable;
use Illuminate\Notifications\Notifiable;
use Illuminate\Support\Collection;

/**
 * @SuppressWarnings(PHPMD.TooManyPublicMethods)
 */
class User extends Model implements ManagedUser
{
    use Authenticatable, CanResetPassword, Notifiable, Authorizable;

    /**
     * @var array
     */
    protected $casts = [
        'approved' => 'boolean',
    ];

    /**
     * @var array
     */
    protected $dates = [
        'login_link_expiry',
    ];

    /**
     * @var string
     */
    protected $table = 'users';

    /**
     * @var array
     */
    protected $fillable = [
        'name',
        'email',
        'role',
        'password',
        'approved',
        'login_link_expiry',
        'first_login_token',
    ];

    /**
     * @param string|array $roles
     */
    public function usersByRoles($roles): Collection
    {
        if (!\is_array($roles)) {
            $roles = [$roles];
        }

        return $this->whereIn('role', $roles)->get();
    }

    public function allAdmins(): Collection
    {
        return $this->usersByRoles(UserRoles::adminRoles());
    }

    public function allSuperAdmins(): Collection
    {
        return $this->usersByRoles(UserRoles::superAdminRoles());
    }

    public function hasAdminAccess(): bool
    {
        return \in_array($this->role, UserRoles::adminRoles(), false);
    }

    public function hasSuperAdminAccess(): bool
    {
        return \in_array($this->role, UserRoles::superAdminRoles(), false);
    }

    public function canManageAdmins(): bool
    {
        return $this->hasSuperAdminAccess();
    }

    public function hasRole(string $role): bool
    {
        return $this->role === $role;
    }

    /**
     * Return if the user has and of the given roles.
     *
     * @param string|array $roles The roles to check
     * @param mixed ...$additionalRoles Unlimited OPTIONAL number of additional roles to check
     */
    public function hasAnyRole($roles, ...$additionalRoles): bool
    {
        return \in_array($this->role, array_merge(array_wrap($roles), $additionalRoles), false);
    }

    public function getByFirstLoginToken(string $token): ManagedUser
    {
        return $this->where('first_login_token', $token)->firstOrFail();
    }

    public function allAwaitingApproval(): Collection
    {
        return static::where('approved', false)->get();
    }

    public function requiresApproving(): bool
    {
        return !$this->approved;
    }

    public function approve(): void
    {
        $this->approved = true;
        $this->save();

        event(new AccountActivated($this));
    }

    public function removeFirstLoginToken(): ManagedUser
    {
        $this->first_login_token = null;

        $this->save();

        return $this;
    }

    public function addLoginToken(string $token, \DateTime $expiry): ManagedUser
    {
        $this->first_login_token = $token;
        $this->login_link_expiry = $expiry;
        $this->save();

        return $this;
    }

    public function loginToken(): ?string
    {
        return $this->first_login_token;
    }

    public function loginExpiry(): ?\DateTime
    {
        return $this->login_link_expiry;
    }
}
